# -*- coding: utf-8 -*-
#
# This file is part of the E2RingSimulator project
#
# Copyright (C): 2019
#                European Synchrotron Radiation Facility
#                BP 220, Grenoble 38043
#                France
# Copyright (C): 2024
#                Elettra - Sinctrone Trieste S.C.p.A.
#
# Distributed under the terms of the LGPL license.
# See LICENSE.txt for more info.

"""
E2RingThread

This class implementes the thread that reads the Elettra 2.0 magnets
und updates the pyAT simulator, the clalculates the new beam parameters
"""

import threading as th
import tango
import timeit
import time
import enum
import numpy as np
import at.tracking
import copy

class Mode(enum.IntEnum):
    """Python enumerated type for Mode attribute."""
    TbT = 0
    Atlinopt = 1
    Atx = 2
    Mixed = 3

class E2RingThread(th.Thread):
    def __init__(self,simulator):
        th.Thread.__init__(self)
        self.simulator=simulator
        self.simulator._counter = 0
        self.detailed_atx_counter = 0 
        self.exc = None # store excpetion of reporting to mother program
    def run(self):
        self.simulator.info_stream("thread started")
        self.running = True;
        self.simulator.set_state(tango.DevState.ON)
        self.simulator.set_status("simulator thread running")
        try:
            while self.running:
                self.simulator.debug_stream("thread loop")
                """
                use the at lattice to calculate the optics parameter and update the simulatro internal values
                """
                tic = timeit.default_timer()
                self.simulator._counter += 1
                if self.simulator._mode == Mode.TbT:
                    self.compute_tbt()
                else:
                    self.compute_optics()
                toc = timeit.default_timer()
                self.simulator._loop_time = toc - tic
                time.sleep(1) # must be tuned or made event dependent...
        except BaseException as e:
            self.exc = e
            se = str(e)
            self.simulator.error_stream("exception in thread %s",se)
            self.simulator.set_status(se)
            self.simulator.set_state(tango.DevState.FAULT)
            return
        self.simulator.info_stream("thread exit")
    def stop(self):
        self.running = False
        
    def join(self):
        th.Thread.join(self)
        if self.exc:
            raise self.exc
        
    def compute_optics(self):
        """
        compute opics paramters and dispatches to sspecilized method depending
        on computation mode
        """
        # init default values
        # self.simulator.debug_stream("compute optic mode %d",self.simulator._mode)
        if self.simulator._mode != Mode.Mixed:
            # self.simulator.debug_stream("mode not mixed")
            self.simulator._emittance_h = np.NaN
            self.simulator._emittance_v = np.NaN
            self.simulator._tune_h = np.NaN
            self.simulator._tune_v = np.NaN
            self.simulator._chromaticity_h = np.NaN
            self.simulator._chromaticity_v = np.NaN
            self.simulator._closed_orbit_h = [np.NaN] * self.simulator._num_bpms
            self.simulator._closed_orbit_v = [np.NaN] * self.simulator._num_bpms
            # self.simulator._closed_orbit_angle_h = [np.NaN] * self.simulator._num_bpms
            # self.simulator._closed_orbit_angle_v = [np.NaN] * self.simulator._num_bpms
            self.simulator._beta_h = [np.NaN] * self.simulator._num_bpms
            self.simulator._beta_v = [np.NaN] * self.simulator._num_bpms
            self.simulator._eta_h = [np.NaN] * self.simulator._num_bpms
            self.simulator._eta_v = [np.NaN] * self.simulator._num_bpms
            #initialize 2D TbT arrays
            # self.simulator.debug_stream("init NaN TbT matrix")
            nanmatrix = np.empty([self.simulator._num_bpms, self.simulator._tb_t__buffer_size])
            nanmatrix[:] = np.NaN
            self.simulator._h_positions_tb_t = nanmatrix
            self.simulator._v_positions_tb_t = nanmatrix
            del nanmatrix

        if self.simulator.ring.radiation: #radiation ON - slower , more detail
            self.simulator.debug_stream("calling find_orbit6")
            _orb_, orb = self.simulator.ring.find_orbit6(refpts=0) 
            l0, beamdata, l = self.simulator.ring.get_optics(refpts=self.simulator.ind_bpm_pinhole, get_chrom=True)
            self.simulator._tune_h = beamdata.tune[0]
            self.simulator._tune_v = beamdata.tune[1]
            self.simulator._chromaticity_h = beamdata.chromaticity[0]
            self.simulator._chromaticity_v = beamdata.chromaticity[1]
            
        else: #radiation OFF - faster, less detail
            self.simulator.debug_stream("calling find_sync_orbit %e", self.simulator.delta_l)
            _orb_, orb = self.simulator.ring.find_sync_orbit(dct=self.simulator.delta_l, refpts=0) #radiation OFF
            l0, t, c, l = self.simulator.ring.linopt(orb[0, 4], refpts=self.simulator.ind_bpm_pinhole, get_chrom=True)
            self.simulator._tune_h = t[0]
            self.simulator._tune_v = t[1]
            self.simulator._chromaticity_h = c[0]
            self.simulator._chromaticity_v = c[1]
            
        lp = l[self.simulator.mask_pin]
        lb = l[self.simulator.mask_bpm]
        
        self.simulator._beta_h = lb['beta'][:, 0]
        self.simulator._beta_v = lb['beta'][:, 1]
        self.simulator._eta_h = lb['dispersion'][:, 0]
        self.simulator._eta_v = lb['dispersion'][:, 2]
        
        self.simulator._closed_orbit_h = lb['closed_orbit'][:,0]
        self.simulator._closed_orbit_v = lb['closed_orbit'][:,2]
        self.simulator._closed_orbit_angle_h = lb['closed_orbit'][:,1]
        self.simulator._closed_orbit_angle_v = lb['closed_orbit'][:,3]
        
        
        # depending on Mode call what is needed for claculating emittance
        if self.simulator._mode == Mode.Atx:
            self.compute_emittance_atx(orb)
            self.detailed_atx_counter += 1
        if self.simulator._mode == Mode.Mixed:
             if not self.simulator._counter % self.simulator._atx_every:
                 self.compute_emittance_atx(orb) # ESRF code check for radiation ON, but compute_emittance_atx turns on radiation anyway...
                 self.detailed_atx_counter += 1

    def compute_emittance_atx(self,orb):
        """
        compute emittances for ring , global and at relevant points
        """
        # need radition ON
        with self.simulator._radlock:
            turn_rad_off = False
            if not self.simulator._radiation:
                self.simulator.ring.radiation_on()
                turn_rad_off = True
                
            # compute what is needed
            emit0, bbb, eee = self.simulator.ring.ohmi_envelope(orbit=orb.T, refpts=self.simulator.ind_pinhole)
            self.simulator._emittance_h = emit0['emitXY'][0]
            self.simulator._emittance_v = emit0['emitXY'][1]
            
            for ip in range(0, len(self.simulator.ind_pinhole)): # extract pinhole data -source points data
                self.simulator.sigmah_pin[ip] = eee['r66'][ip][0][0] 
                self.simulator.sigmav_pin[ip] = eee['r66'][ip][2][2]
                self.simulator.sigmahv_pin[ip] = eee['r66'][ip][2][0]
         
                self.simulator.posh_pin[ip] = self.simulator._closed_orbit_h[ip] # extract positions and angle pinhole - source points
                self.simulator.posv_pin[ip] = self.simulator._closed_orbit_v[ip] 
                self.simulator.angleh_pin[ip] = self.simulator._closed_orbit_angle_h[ip] 
                self.simulator.anglev_pin[ip] = self.simulator._closed_orbit_angle_v[ip] 
                
            
            if turn_rad_off:
                self.simulator.ring.radiation_off()
                
    def compute_tbt(self):
        incoord = copy.deepcopy(self.simulator._tb_t__in_coord) #make a deep copy becausse  lattice_pass modifies the inpout coordinate
        try:
            coord = at.lattice_pass(
                self.simulator.ring,
                incoord,
                nturns=self.simulator._tb_t__buffer_size,
                refpts=self.simulator.ind_bpms)
        except Exception as e:
            coord = np.array(list(self.simulator._tb_t__in_coord))
            self.exc = e
            se = str(e)
            self.simulator.error_stream("compute_tbt: %s", se)
        self.simulator._h_positions_tb_t = coord[0, 0, :, :] # get X
        self.simulator._v_positions_tb_t = coord[3, 0, :, :] # get Y
        # TODO: add losses flga and extrct lossess parameters . now attribute for reading losses
        # TODO: creat bunch end follow losses - calculation heavy!
        
