# -*- coding: utf-8 -*-
#
# This file is part of the E2RingSimulator project
#
# Copyright (C): 2019
#                European Synchrotron Radiation Facility
#                BP 220, Grenoble 38043
#                France
# Copyright (C): 2024
#                Elettra - Sinctrone Trieste S.C.p.A.
#
# Distributed under the terms of the LGPL license.
# See LICENSE.txt for more info.

"""
E2RingThread

This class implementes the thread that updates  the Elettra 2.0 pyAt simulator, and calculates the new beam parameters
"""

import threading as th
import tango
import timeit
import time
import enum
import numpy as np
import at.tracking
import copy


class E2RingUpdateThread(th.Thread):

    def __init__(self, simulator):
        th.Thread.__init__(self)
        self.simulator = simulator  # keep reference to simulator for accessing data
        self.exc = None  # store excpetion of reporting to mother program
        self.running = False
        self.changed = False

    def run(self):
        self.running = True
        while self.running:
            tic = timeit.default_timer()
            # do somethingtimeit.default_timer()
            self.read_devices()
            if self.changed:
                # print("mags changed ====>", self.changed)
                self.set_ring()
            toc = timeit.default_timer()
            self.simulator._update_time = toc - tic
            #print(self.loop_time) TODO: add read/update time attribute in E2ringsimulator for monitoring
            time.sleep(1)  # must be tuned or made event dependent...

    def stop(self):
        self.running = False
        
    def join(self):
        th.Thread.join(self)
        if self.exc:
            raise self.exc
        
    def read_devices(self):
        """
        read_devices(self)
        core method of the thread: scan the lists of ElementProxies and read the strengths
        """
        
        # quadrupoles
        qchanged = False
        for qeleproxy in self.simulator.qeleproxy:
            qeleproxy.read()
            if qeleproxy.valid and qeleproxy.changed:
                qchanged = True
        
        #sextupoles
        schanged = False
        for seleproxy in self.simulator.seleproxy:
            seleproxy.read()
            if seleproxy.valid and seleproxy.changed:
                schanged = True
        
        #corrector
        cchanged = False
        for celeproxy in self.simulator.celeproxy:
            celeproxy.read()
            if celeproxy.valid and celeproxy.changed:
                cchanged = True
        self.changed = qchanged or schanged or cchanged
        
    def set_ring(self):
        """
        scan element proxy lists and sets changed strength to ring elements
        TODO: set values for good!
        """
        for qeleproxy in self.simulator.qeleproxy:  # QUADRUPOLES
            if qeleproxy.valid and qeleproxy.changed:
                idx = qeleproxy.mag_index
                newK = qeleproxy.strength/self.simulator.ring[idx].Length
                self.simulator.ring[idx].K = newK
                #print(self.simulator.ring[idx].FamName,qeleproxy.mag_name,newK, self.simulator.ring[idx].K)
            
        for seleproxy in self.simulator.seleproxy:  # SEXTUPOLES
            if seleproxy.valid and seleproxy.changed:
                idx = seleproxy.mag_index
                newH = seleproxy.strength/self.simulator.ring[idx].Length
                self.simulator.ring[idx].H = newH
                #print(self.simulator.ring[idx].FamName,seleproxy.mag_name,newH, self.simulator.ring[idx].H)
                
                #  TODO: OCTUPOLES
                
        for celeproxy in self.simulator.celeproxy:  # CORRECTORS
            if celeproxy.valid and celeproxy.changed:
                idx = celeproxy.mag_index
                newC = celeproxy.strength
                hv = celeproxy.hv
                if hv == None:
                    print("MALCONFIGURATO!", celeproxy.mag_nam)
                else:
                    self.simulator.ring[idx].KickAngle[hv] = newC
                    # print(self.simulator.ring[idx].FamName,celeproxy.mag_name,newC, self.simulator.ring[idx].KickAngle[hv],hv,self.simulator.ring[idx].PassMethod)
            
